from .. import context

from agilicus import agilicus_api

from ..input_helpers import (
    build_updated_model,
    model_from_dict,
    update_org_from_input_or_ctx,
)
from ..input_helpers import strip_none
from ..output.table import (
    column,
    spec_column,
    format_table,
    metadata_column,
    object_subtable,
)

from .formatters import constraint_columns
from ..licensing.product_table_versions import list_product_table_versions


def list_licenses(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    params = strip_none(kwargs)
    query_results = apiclient.licensing_api.list_licenses(**params)
    return query_results.licenses


def add_license(
    ctx, product_table_version, product_name, constraints=None, vars=None, **kwargs
):
    apiclient = context.get_apiclient_from_ctx(ctx)
    spec = agilicus_api.LicenseSpec(
        product_table_version=agilicus_api.ProductTableVersionString(
            product_table_version
        ),
        product_name=agilicus_api.LicensedProductName(product_name),
        **strip_none(kwargs),
    )
    if constraints:
        spec.license_constraints = build_constraints(constraints)
    if vars:
        spec.constraint_variables = build_vars(vars)

    model = agilicus_api.License(spec=spec)

    return apiclient.licensing_api.create_license(model)


def _add_constraint_defaults(**kwargs):
    if "priority" not in kwargs:
        kwargs["priority"] = 0
    return kwargs


def build_constraints(constraints):
    return [
        model_from_dict(
            agilicus_api.LicenseConstraint, _add_constraint_defaults(**constraint)
        )
        for constraint in constraints
    ]


def build_vars(vars):
    return agilicus_api.LicenseConstraintVariables._from_openapi_data(**vars)


def _get_license(ctx, apiclient, license_id, **kwargs):
    return apiclient.licensing_api.get_license(license_id, **kwargs)


def show_license(ctx, license_id, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    return _get_license(ctx, apiclient, license_id, **kwargs)


def delete_license(ctx, license_id, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    return apiclient.licensing_api.delete_license(license_id, **kwargs)


def update_license(
    ctx,
    license_id,
    constraints=None,
    vars=None,
    replace_constraints=None,
    replace_vars=None,
    subscription_reconcile=None,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)
    mapping = _get_license(ctx, apiclient, license_id)

    apply_constraint_and_vars(
        mapping.spec, constraints, vars, replace_constraints, replace_vars
    )

    # check_type=False works around nested types not deserializing correctly
    mapping.spec = build_updated_model(
        agilicus_api.LicenseSpec, mapping.spec, kwargs, check_type=False
    )

    query = {}
    if subscription_reconcile:
        query["subscription_reconcile"] = subscription_reconcile
    return apiclient.licensing_api.replace_license(license_id, license=mapping, **query)


def apply_constraint_and_vars(obj, constraints, vars, replace_constraints, replace_vars):
    if not constraints:
        constraints = {}
    if not vars:
        vars = {}

    if not replace_constraints:
        constraints = build_constraints(constraints) + (obj.license_constraints or [])
    else:
        constraints = build_constraints(constraints)

    if not replace_vars and obj.constraint_variables:
        vars = obj.constraint_variables.to_dict() | vars
    vars = build_vars(vars)

    obj.license_constraints = constraints
    obj.constraint_variables = vars


def format_licenses(ctx, resources):
    columns = [
        metadata_column("id"),
        spec_column("product_table_version"),
        spec_column("product_name"),
        *constraint_columns(10, column_func=spec_column),
    ]

    return format_table(ctx, resources, columns)


def list_license_details(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    kwargs = strip_none(kwargs)

    result = apiclient.licensing_api.list_license_details(**kwargs)
    return result.license_details


def format_license_details(ctx, resources):
    columns = [
        column("org_id"),
        column("license.metadata.id", "license"),
        column("license.spec.product_name", "product_name"),
        column("license.spec.product_table_version", "product_table_version"),
        *constraint_columns(10),
    ]

    return format_table(ctx, resources, columns)


def list_license_evaluation_contexts(ctx, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    update_org_from_input_or_ctx(kwargs, ctx, **kwargs)
    kwargs = strip_none(kwargs)

    result = apiclient.licensing_api.list_license_evaluation_contexts(**kwargs)
    return result.license_evaluation_contexts


def format_license_evaluation_contexts(ctx, resources):
    usage_columns = [
        column("num_resources"),
        column("num_desktops"),
        column("num_applications"),
        column("num_networks"),
        column("num_ssh"),
        column("num_databases"),
        column("num_resource_groups"),
        column("num_fileshares"),
        column("num_launchers"),
        column("num_users"),
        column("num_groups"),
        column("num_orgs"),
        column("num_connectors"),
    ]
    columns = [
        column("org_id"),
        column("subscription_id"),
        object_subtable(
            ctx, "license_evaluation_input.subscription.usage", usage_columns, "usage"
        ),
    ]

    return format_table(ctx, resources, columns)


def get_licensed_product(product_label, product_table):
    if not product_label:
        return None

    product_tokens = product_label.split("-")

    if len(product_tokens) == 1:
        normalized_product_name = "Standard"
    elif len(product_tokens) >= 2:
        # everything but the last, which is the currency
        normalized_product_name = "-".join(product_tokens[:-1])

    for licensed_product in product_table["products"]:
        if (
            str(licensed_product["name"]).casefold()
            == normalized_product_name.casefold()
        ):
            return licensed_product["name"]
    return None


def add_license_to_billing_sub(ctx, bsub, product_name=None, product_table_version=None):
    if bsub.spec.license_id:
        raise Exception("subscription already has a license_id set!")

    product_table = list_product_table_versions(ctx, published=True, limit=1)
    product_version = product_table[0]

    if not product_table_version:
        product_table_version = product_version["spec"]["version"]

    product_table = product_version["spec"].get("product_table")
    if not product_name:
        if not bsub.status.product:
            raise Exception(f"no product for subscription {bsub.metadata.id}")

        product_name = get_licensed_product(
            bsub.status.product.spec.label, product_table
        )
        if not product_name:
            raise Exception(
                "could not find a suitable licensed for product label: "
                f"{bsub.status.product.spec.label}"
            )

    result = add_license(
        ctx,
        product_table_version=str(product_table_version),
        product_name=str(product_name),
    )
    bsub.spec.license_id = result.metadata.id




# flake8: noqa

# Import all APIs into this package.
# If you have many APIs here with many many models used in each API this may
# raise a `RecursionError`.
# In order to avoid this, import only the API that you directly need like:
#
#   from .api.application_services_api import ApplicationServicesApi
#
# or import this package, but before doing it, use:
#
#   import sys
#   sys.setrecursionlimit(n)

# Import APIs into API package:
from agilicus_api.api.application_services_api import ApplicationServicesApi
from agilicus_api.api.applications_api import ApplicationsApi
from agilicus_api.api.audits_api import AuditsApi
from agilicus_api.api.billing_api import BillingApi
from agilicus_api.api.catalogues_api import CataloguesApi
from agilicus_api.api.certificates_api import CertificatesApi
from agilicus_api.api.challenges_api import ChallengesApi
from agilicus_api.api.connectors_api import ConnectorsApi
from agilicus_api.api.credentials_api import CredentialsApi
from agilicus_api.api.deployments_api import DeploymentsApi
from agilicus_api.api.diagnostics_api import DiagnosticsApi
from agilicus_api.api.features_api import FeaturesApi
from agilicus_api.api.files_api import FilesApi
from agilicus_api.api.groups_api import GroupsApi
from agilicus_api.api.hosts_api import HostsApi
from agilicus_api.api.issuers_api import IssuersApi
from agilicus_api.api.labels_api import LabelsApi
from agilicus_api.api.launchers_api import LaunchersApi
from agilicus_api.api.licensing_api import LicensingApi
from agilicus_api.api.lookups_api import LookupsApi
from agilicus_api.api.messages_api import MessagesApi
from agilicus_api.api.metrics_api import MetricsApi
from agilicus_api.api.organisations_api import OrganisationsApi
from agilicus_api.api.permissions_api import PermissionsApi
from agilicus_api.api.policy_api import PolicyApi
from agilicus_api.api.policy_templates_api import PolicyTemplatesApi
from agilicus_api.api.policy_config_api import PolicyConfigApi
from agilicus_api.api.regions_api import RegionsApi
from agilicus_api.api.resources_api import ResourcesApi
from agilicus_api.api.rules_api import RulesApi
from agilicus_api.api.tokens_api import TokensApi
from agilicus_api.api.trusted_certs_api import TrustedCertsApi
from agilicus_api.api.users_api import UsersApi
from agilicus_api.api.whoami_api import WhoamiApi



from copy import deepcopy
from . import context
from . import input_helpers
from .input_helpers import get_org_from_input_or_ctx, build_alternate_mode_setting
from .output.table import (
    format_table,
    metadata_column,
    spec_column,
    column,
    subtable,
)
from .pagination import normalize_page_args
from .resource_helpers import standard_page_fields

import agilicus

page_fields = standard_page_fields


def get(ctx, id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    return apiclient.launchers_api.get_launcher(id, org_id=org_id, **kwargs)


def delete(ctx, id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    return apiclient.launchers_api.delete_launcher(id, org_id=org_id, **kwargs)


def add(ctx, org_id=None, resource_members=[], **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    kwargs.pop("org_id", None)
    name = kwargs.pop("name")

    spec = agilicus.LauncherSpec(org_id=org_id, name=name)
    for k, v in deepcopy(kwargs).items():
        if v is None:
            kwargs.pop(k)

    spec.config = agilicus.LauncherConfig(**kwargs)
    member_objs = []
    for member in resource_members:
        member_objs.append(agilicus.ResourceMember(id=member))
    spec.resource_members = member_objs

    alternate_mode_setting = build_alternate_mode_setting(
        None,
        learning_mode=kwargs.get("learning_mode", None),
        learning_mode_expiry=kwargs.get("learning_mode_expiry", None),
        diagnostic_mode=kwargs.get("diagnostic_mode", None),
    )
    spec.alternate_mode_setting = alternate_mode_setting

    launcher = agilicus.Launcher(spec=spec)
    return apiclient.launchers_api.create_launcher(launcher)


def query(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs = input_helpers.strip_none(kwargs)

    kwargs["org_id"] = org_id
    kwargs = normalize_page_args(kwargs)
    query_results = apiclient.launchers_api.list_launchers(**kwargs)
    return query_results.launchers


def replace(  # noqa
    ctx,
    id,
    resource_members=None,
    remove_resource_members=None,
    name=None,
    org_id=None,
    command_path=None,
    command_arguments=None,
    start_in=None,
    do_intercept=None,
    hide_console=None,
    disable_http_proxy=None,
    application_ids=None,
    remove_application_ids=None,
    run_as_admin=None,
    fork_then_attach=None,
    end_existing_if_running=None,
    wait_for_all_descendants=None,
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    resource = apiclient.launchers_api.get_launcher(id, org_id=org_id)
    if remove_resource_members is not None:
        old_members = resource.spec.resource_members
        resource.spec.resource_members = []
        for member in old_members:
            if member.id in remove_resource_members:
                # needs to be removed.
                continue
            resource.spec.resource_members.append(member)
    if resource_members is not None:
        for member in resource_members:
            resource.spec.resource_members.append(agilicus.ResourceMember(id=member))

    if remove_application_ids is not None:
        old_ids = resource.spec.applications
        resource.spec.applications = []
        for app_id in old_ids:
            if app_id in remove_application_ids:
                # needs to be removed.
                continue
            resource.spec.applications.append(app_id)
    if application_ids is not None:
        for app_id in application_ids:
            resource.spec.applications.append(app_id)

    if name is not None:
        resource.spec.name = name
    if resource.spec.config is None:
        resource.spec.config = agilicus.LauncherConfig()
    if command_path is not None:
        resource.spec.config.command_path = command_path
    if command_arguments is not None:
        resource.spec.config.command_arguments = command_arguments
    if start_in is not None:
        resource.spec.config.start_in = start_in
    if do_intercept is not None:
        resource.spec.config.do_intercept = do_intercept
    if hide_console is not None:
        resource.spec.config.hide_console = hide_console
    if disable_http_proxy is not None:
        resource.spec.config.disable_http_proxy = disable_http_proxy
    if run_as_admin is not None:
        resource.spec.config.run_as_admin = run_as_admin
    if fork_then_attach is not None:
        if resource.spec.config.interceptor_config is None:
            resource.spec.config.interceptor_config = agilicus.InterceptorConfig()
        resource.spec.config.interceptor_config.fork_then_attach = fork_then_attach
    if end_existing_if_running is not None:
        resource.spec.config.end_existing_if_running = end_existing_if_running
    if wait_for_all_descendants is not None:
        resource.spec.config.wait_for_all_descendants = wait_for_all_descendants

    resource.spec.alternate_mode_setting = build_alternate_mode_setting(
        resource.spec.alternate_mode_setting,
        learning_mode=kwargs.get("learning_mode"),
        learning_mode_expiry=kwargs.get("learning_mode_expiry"),
        diagnostic_mode=kwargs.get("diagnostic_mode", None),
    )

    return apiclient.launchers_api.replace_launcher(id, launcher=resource).to_dict()


def add_interceptor_rule(
    ctx,
    id,
    org_id=None,
    allow_name_exact_list=[],
    allow_value_regex_list=[],
    disallow_name_exact_list=[],
    disallow_value_regex_list=[],
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    allow_commands = []
    for name in allow_name_exact_list:
        allow_commands.append(agilicus.InterceptorCommand(name_exact=name))

    for regex in allow_value_regex_list:
        allow_commands.append(agilicus.InterceptorCommand(value_regex=regex))

    disallow_commands = []
    for name in disallow_name_exact_list:
        disallow_commands.append(agilicus.InterceptorCommand(name_exact=name))

    for regex in disallow_value_regex_list:
        disallow_commands.append(agilicus.InterceptorCommand(value_regex=regex))

    resource = apiclient.launchers_api.get_launcher(id, org_id=org_id)

    _initialize_interceptor_config(resource)

    resource.spec.config.interceptor_config.allow_list.extend(allow_commands)
    resource.spec.config.interceptor_config.disallow_list.extend(disallow_commands)

    return apiclient.launchers_api.replace_launcher(id, launcher=resource).to_dict()


def add_interceptor_extra_process(
    ctx,
    id,
    program_name,
    org_id=None,
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    resource = apiclient.launchers_api.get_launcher(id, org_id=org_id)

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    new_extra_program = agilicus.ExtraProcess(program_name=program_name, **kwargs)

    program_list = resource.spec.config.extra_processes
    if program_list is None:
        program_list = []
        resource.spec.config.extra_processes = program_list

    program_list.append(new_extra_program)

    return apiclient.launchers_api.replace_launcher(id, launcher=resource).to_dict()


def remove_interceptor_extra_process(
    ctx,
    id,
    program_name,
    org_id=None,
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    resource = apiclient.launchers_api.get_launcher(id, org_id=org_id)

    program_list = resource.spec.config.extra_processes
    if not program_list:
        print("no extra process to remove")
        return

    item_number = -1
    for index, prog in enumerate(program_list):
        if prog.program_name == program_name:
            item_number = index
            break

    if item_number == -1:
        print(f"extra process {program_name} not found")
        return

    program_list.pop(item_number)
    return apiclient.launchers_api.replace_launcher(id, launcher=resource).to_dict()


def _initialize_interceptor_config(launcher):
    if not launcher.spec.config.interceptor_config:
        launcher.spec.config.interceptor_config = agilicus.InterceptorConfig()

    if not launcher.spec.config.interceptor_config.allow_list:
        launcher.spec.config.interceptor_config.allow_list = []

    if not launcher.spec.config.interceptor_config.disallow_list:
        launcher.spec.config.interceptor_config.disallow_list = []


def remove_name(interceptor_command_list, name):
    new_list = []
    for command in interceptor_command_list:
        if command.get("name_exact") != name:
            new_list.append(command)
    return new_list


def remove_regex(interceptor_command_list, regex):
    new_list = []
    for command in interceptor_command_list:
        if command.get("value_regex") != regex:
            new_list.append(command)
    return new_list


def remove_interceptor_rule(
    ctx,
    id,
    org_id=None,
    allow_name_exact_list=[],
    allow_value_regex_list=[],
    disallow_name_exact_list=[],
    disallow_value_regex_list=[],
    **kwargs,
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    resource = apiclient.launchers_api.get_launcher(id, org_id=org_id)

    _initialize_interceptor_config(resource)

    for name in allow_name_exact_list:
        resource.spec.config.interceptor_config.allow_list = remove_name(
            resource.spec.config.interceptor_config.allow_list, name
        )

    for regex in allow_value_regex_list:
        resource.spec.config.interceptor_config.allow_list = remove_regex(
            resource.spec.config.interceptor_config.allow_list, regex
        )

    for name in disallow_name_exact_list:
        resource.spec.config.interceptor_config.disallow_list = remove_name(
            resource.spec.config.interceptor_config.disallow_list, name
        )

    for regex in disallow_value_regex_list:
        resource.spec.config.interceptor_config.disallow_list = remove_regex(
            resource.spec.config.interceptor_config.disallow_list, regex
        )

    return apiclient.launchers_api.replace_launcher(id, launcher=resource).to_dict()


def format_launchers(ctx, launchers):
    app_service_columns = [
        column("id"),
        column("name"),
    ]
    fs_columns = [
        metadata_column("id"),
        spec_column("name"),
    ]
    columns = [
        metadata_column("id"),
        spec_column("org_id", "org id"),
        spec_column("name", "name"),
        spec_column("config", "config"),
        spec_column("applications"),
        subtable(ctx, "status.file_shares", fs_columns, optional=True),
        subtable(ctx, "status.application_services", app_service_columns, optional=True),
    ]
    return format_table(ctx, launchers, columns)



import click

from ..output.table import output_entry

from agilicus.command_helpers import Command

from . import databases
from ..input_helpers import page_sort_order_values, search_direction_values

cmd = Command()


@cmd.command(name="list-database-resources")
@click.option("--name", default=None)
@click.option("--org-id", default=None)
@click.option("--connector-id", default=None)
@click.option("--page-at-id", default=None)
@click.option("--name-slug", default=None)
@click.option("--updated-since", default=None, type=click.DateTime())
@click.option("--resource-id", default=None)
@click.option("--limit", default=500)
@click.option(
    "--page-on", multiple=True, type=click.Choice(databases.page_fields), default=None
)
@click.option("--page-at-key", multiple=True, type=str, default=None)
@click.option(
    "--page-sort", multiple=True, type=click.Choice(page_sort_order_values), default=None
)
@click.option(
    "--search-direction", type=click.Choice(search_direction_values), default=None
)
@click.pass_context
def list_database_resources(ctx, name=None, **kwargs):
    resources = databases.list_database_resources(ctx, name=name, **kwargs)
    table = databases.format_database_as_text(ctx, resources)
    print(table)


@cmd.command(name="add-database-resource")
@click.option("--name", default=None, required=True)
@click.option("--org-id", default=None)
@click.option("--connector-id", default=None)
@click.option("--address", default=None, required=True)
@click.option("--port", type=int, default=None, required=True)
@click.option("--data-source-name", type=str)
@click.option("--runtime-parameter", type=(str, str), multiple=True)
@click.option(
    "--database-protocol", type=click.Choice(databases.PROTOCOLS), required=True
)
@click.pass_context
def add_database_resource(ctx, name, port, runtime_parameter, **kwargs):
    result = databases.add_database_resource(
        ctx, name=name, port=port, runtime_parameters=runtime_parameter, **kwargs
    )
    output_entry(ctx, result)


@cmd.command(name="update-database-resource")
@click.argument("database-resource-id")
@click.option("--name", default=None)
@click.option("--org-id", default=None)
@click.option("--connector-id", default=None)
@click.option("--address", default=None)
@click.option("--port", type=int, default=None)
@click.option("--username", default=None)
@click.option("--published", type=click.Choice(["no", "public"]), default=None)
@click.option("--runtime-parameter", type=str, multiple=True)
@click.option(
    "--replace-runtime-parameters",
    is_flag=True,
    help="if set, replaces runtime parameters with given values",
)
@click.option("--data-source-name", type=str, default=None)
@click.option(
    "--database-protocol", type=click.Choice(databases.PROTOCOLS), default=None
)
@click.pass_context
def update_database_resource(
    ctx, database_resource_id, port, published, runtime_parameter, **kwargs
):
    result = databases.update_database_resource(
        ctx,
        database_resource_id,
        port,
        published,
        runtime_parameters=runtime_parameter,
        **kwargs,
    )
    output_entry(ctx, result)


@cmd.command(name="show-database-resource")
@click.argument("database-resource-id")
@click.option("--org-id", default=None)
@click.pass_context
def show_database_resource(ctx, database_resource_id, **kwargs):
    result = databases.show_database_resource(ctx, database_resource_id, **kwargs)
    output_entry(ctx, result)


@cmd.command(name="delete-database-resource")
@click.argument("database-resource-id")
@click.option("--org-id", default=None)
@click.pass_context
def delete_database_resource(ctx, database_resource_id, **kwargs):
    databases.delete_database_resource(ctx, database_resource_id, **kwargs)


def add_commands(cli):
    cmd.add_to_cli(cli)



import pytest
from unittest.mock import Mock
from agilicus.pagination.auto_iterator import (
    AutoIterator,
    AutoIteratorWrapper,
    get_page_on,
    _determine_page_on,
    patch_endpoint,
    _handle_page_failure,
)
from agilicus_api.exceptions import ApiException
from agilicus import ListSSHResourcesResponse
from agilicus import PageAtKey
from agilicus_api.api_client import Endpoint


class MockPageInfo:
    def __init__(self, page_at_key, page_on_value):
        self.page_at_key = page_at_key
        self.page_on = Mock(value=page_on_value)


class MockResponse:
    def __init__(
        self, items, page_at_key=None, page_on_value=None, legacy_pagination=False
    ):
        self.items = items
        self.attribute_map = {"items": "list[object]"}
        self.page_info = (
            MockPageInfo(page_at_key, page_on_value)
            if page_at_key and page_on_value
            else None
        )
        self.legacy_page_info = {}
        if legacy_pagination and page_on_value and page_at_key:
            self.legacy_page_info["page_at_" + page_on_value[0]] = page_at_key[0]

    def get(self, key, default=None):
        if key == "items":
            return self.items
        if self.legacy_page_info:
            return self.legacy_page_info.get(key, default)
        return default


class MockItem:
    def __init__(self, name, index):
        self.name = name
        self.id = index


class MockListFunction:
    def __init__(
        self,
        items_per_page,
        total_items,
        page_on_allowed=None,
        legacy_pagination_key=None,
    ):
        self._all_items = [MockItem(f"item_{i}", i) for i in range(total_items)]
        self._items_per_page = items_per_page
        self.call_count = 0
        self.allowed_values = {}
        self.params_map = {"all": []}
        if page_on_allowed is None:
            page_on_allowed = ["id"]
        if page_on_allowed:
            self.allowed_values = {("page_on",): {k: k for k in page_on_allowed}}
            self.params_map["all"].append("page_on")
        self.legacy_pagination_key = legacy_pagination_key

    def __call__(
        self, page_on=None, page_at_key=None, limit=None, page_sort=None, **kwargs
    ):
        if page_at_key is not None and not isinstance(page_at_key, list):
            raise ValueError("page_at_key must be a list")

        if limit is None:
            limit = 1000
        if page_on is None:
            page_on = []
        self.call_count += 1
        start_index = 0
        if self.legacy_pagination_key:
            page_on = [self.legacy_pagination_key]
            page_at_key = [kwargs["page_at_" + self.legacy_pagination_key]]
        if page_at_key and page_at_key[0] not in (None, ""):
            page_at_key[0] = int(page_at_key[0])
            try:
                start_index = len(self._all_items)
                for idx, item in enumerate(self._all_items):
                    compare_tuple = [getattr(item, key) for key in page_on]
                    if page_at_key < compare_tuple:
                        start_index = idx
                        break
            except ValueError:
                start_index = len(self._all_items)

        end_index = min(start_index + limit, len(self._all_items))
        current_items = self._all_items[start_index:end_index]

        next_page_at_key = None
        if end_index < len(self._all_items):
            key = [str(self._all_items[end_index - 1].id)]
            if self.legacy_pagination_key:
                next_page_at_key = key
            else:
                next_page_at_key = PageAtKey(key)

        return MockResponse(
            current_items,
            page_at_key=next_page_at_key,
            page_on_value=page_on if page_on else None,
            legacy_pagination=self.legacy_pagination_key is not None,
        )


def test_auto_iterator_basic_pagination():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    items = list(iterator)
    assert len(items) == 5
    assert mock_list_function.call_count == 3  # 5 items, 2 per page = 3 calls (2,2,1)
    assert [item.name for item in items] == [
        "item_0",
        "item_1",
        "item_2",
        "item_3",
        "item_4",
    ]


def test_auto_iterator_basic_legacy():
    mock_list_function = MockListFunction(
        items_per_page=2, total_items=5, legacy_pagination_key="id"
    )
    iterator = AutoIterator(
        mock_list_function,
        limit=None,
        page_size=2,
        page_sort=[],
        page_on="page_at_id",
        legacy_pagination=True,
    )

    items = list(iterator)
    assert len(items) == 5
    assert mock_list_function.call_count == 3  # 5 items, 2 per page = 3 calls (2,2,1)
    assert [item.name for item in items] == [
        "item_0",
        "item_1",
        "item_2",
        "item_3",
        "item_4",
    ]


def test_auto_iterator_with_limit():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)
    iterator = AutoIterator(mock_list_function, limit=3, page_size=2, page_sort=[])

    items = list(iterator)
    assert len(items) == 3
    assert mock_list_function.call_count == 2  # 3 items, 2 per page = 2 calls (2,1)
    assert [item.name for item in items] == ["item_0", "item_1", "item_2"]


def test_auto_iterator_empty_result():
    mock_list_function = MockListFunction(items_per_page=2, total_items=0)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    items = list(iterator)
    assert len(items) == 0
    assert mock_list_function.call_count == 1


def test_auto_iterator_first_method():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    first_item = iterator.first()
    assert first_item.name == "item_0"
    assert mock_list_function.call_count == 1

    # Calling first again should re-initialize and get the first item again
    first_item = iterator.first()
    assert first_item.name == "item_0"
    assert mock_list_function.call_count == 2


def test_auto_iterator_first_method_empty():
    mock_list_function = MockListFunction(items_per_page=2, total_items=0)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    first_item = iterator.first()
    assert first_item is None
    assert mock_list_function.call_count == 1


def test_auto_iterator_all_method():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    all_items = iterator.all()
    assert len(all_items) == 5
    assert mock_list_function.call_count == 3
    assert [item.name for item in all_items] == [
        "item_0",
        "item_1",
        "item_2",
        "item_3",
        "item_4",
    ]


def test_auto_iterator_all_method_empty():
    mock_list_function = MockListFunction(items_per_page=2, total_items=0)
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    all_items = iterator.all()
    assert len(all_items) == 0
    assert mock_list_function.call_count == 1


def test_auto_iterator_wrapper():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)
    wrapper = AutoIteratorWrapper(mock_list_function)
    iterator = wrapper(limit=3, page_size=2, page_sort=["name"])

    assert isinstance(iterator, AutoIterator)
    assert iterator.limit == 3
    assert iterator.page_size == 2
    assert iterator.page_sort == ["name"]


def test_get_page_on_with_allowed_values():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {("page_on",): {"id": "id", "name": "name"}}
    mock_list_function.params_map = {"all": ["page_on"]}
    assert get_page_on(mock_list_function) == ["id", "name"]


def test_get_page_on_without_allowed_values_but_in_params_map():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {}
    mock_list_function.params_map = {"all": ["page_on"]}
    assert get_page_on(mock_list_function) == ["id"]


def test_get_page_on_not_in_params_map():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {}
    mock_list_function.params_map = {"all": []}
    assert get_page_on(mock_list_function) == []


def test_determine_page_on_preference():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {
        ("page_on",): {"created": "created", "id": "id"}
    }
    mock_list_function.params_map = {"all": ["page_on"]}
    assert _determine_page_on(mock_list_function) == ["created"]


def test_determine_page_on_default():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {("page_on",): {"foo": "foo"}}
    mock_list_function.params_map = {"all": ["page_on"]}
    assert _determine_page_on(mock_list_function) == ["foo"]


def test_determine_page_on_not_implemented():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {}
    mock_list_function.params_map = {"all": []}
    with pytest.raises(NotImplementedError):
        _determine_page_on(mock_list_function)


def test_patch_endpoint():
    mock_list_function = Mock()
    patch_endpoint(mock_list_function)
    assert hasattr(mock_list_function, "auto_paging_iter")
    assert isinstance(mock_list_function.auto_paging_iter, AutoIteratorWrapper)


def test_handle_page_failure_pagination_key_error():
    exc = ApiException(status=400, reason="Bad Request", http_resp=Mock())
    exc.body = '{"error_message": "Invalid pagination key"}'
    with pytest.raises(
        NotImplementedError, match="Pagination not implemented on this key"
    ):
        _handle_page_failure(exc.body)


def test_handle_page_failure_other_400_error():
    exc = ApiException(status=400, reason="Bad Request", http_resp=Mock())
    exc.body = '{"error_message": "Some other error"}'
    with pytest.raises(ApiException):
        try:
            raise exc
        except Exception as exc:
            _handle_page_failure(exc.body)


def test_handle_page_failure_no_body():
    exc = ApiException(status=400, reason="Bad Request", http_resp=Mock())
    exc.body = None
    with pytest.raises(ApiException):
        try:
            raise exc
        except Exception as exc:
            _handle_page_failure(exc.body)


def test_handle_page_failure_invalid_json():
    exc = ApiException(status=400, reason="Bad Request", http_resp=Mock())
    exc.body = "invalid json"
    with pytest.raises(ApiException):
        try:
            raise exc
        except Exception as exc:
            _handle_page_failure(exc.body)


def test_auto_iterator_api_exception_other_status():
    mock_list_function = Mock()
    mock_list_function.allowed_values = {("page_on",): {"foo": "foo"}}
    mock_list_function.params_map = {"all": ["page_on"]}
    mock_list_function.side_effect = ApiException(
        status=500, reason="Internal Server Error"
    )
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])

    with pytest.raises(ApiException):
        list(iterator)


def test_auto_iterator_repr():
    mock_list_function = Mock(
        __name__="mock_list_function",
        allowed_values={("page_on",): {"id": "id"}},
        params_map={"all": ["page_on"]},
    )
    iterator = AutoIterator(mock_list_function, limit=None, page_size=2, page_sort=[])
    assert repr(iterator) == "<AutoIterator for mock_list_function>"


def test_auto_patch_endpoint():
    mock_list_function = MockListFunction(items_per_page=2, total_items=5)

    def wrap_list(*args, **kwargs):
        return mock_list_function(**kwargs)

    foo = Endpoint(
        settings={
            "response_type": (ListSSHResourcesResponse,),
            "auth": ["token-valid"],
            "endpoint_path": "/v1/ssh_resources",
            "operation_id": "list_ssh_resources",
            "http_method": "GET",
            "servers": None,
        },
        params_map={
            "all": [
                "page_on",
                "page_at_key",
                "page_sort",
                "search_direction",
            ],
            "required": [],
            "nullable": [],
            "enum": [
                "page_sort",
                "search_direction",
            ],
            "validation": [
                "name",
                "limit",
            ],
        },
        root_map={
            "validations": [],
            "allowed_values": {("page_on,"): ["name"]},
            "openapi_types": {},
            "attribute_map": {},
            "location_map": {},
            "collection_format_map": {},
        },
        callable=wrap_list,
    )

    assert foo.auto_paging_iter is not None
    assert next(foo.auto_paging_iter()).name == "item_0"



import datetime
from datetime import timedelta
import agilicus
from prettytable import PrettyTable
from operator import attrgetter
from . import get_many_entries

import agilicus_api.exceptions
import dateutil.tz
import operator
import urllib.parse
from colorama import Fore
import re
from typing import Optional
from urllib.parse import urlparse

from . import context
from . import output
from . import regions
from .input_helpers import get_org_from_input_or_ctx
from .input_helpers import strip_none
from agilicus import input_helpers
from .orgs import get_org_by_dictionary
from . import create_or_update
from .custom_types import Ternary
from . import pagination
from .resource_helpers import standard_page_fields

from .output.table import (
    spec_column,
    status_column,
    format_table,
    column,
    metadata_column,
    subtable,
)

TUNNEL_TERMINATION_TYPES = ["tcp", "inproc"]
INTERNAL_SERVICE_BIND = ["disabled", "local", "all", "custom"]

page_fields = standard_page_fields


def _filter_version(connector, not_version=None, only_version=None):
    if not not_version and not only_version:
        return False

    filter_out = True
    for instance in connector["status"].get("instances", []):
        try:
            v = _get_version(instance, "")
            if not_version and v != not_version:
                filter_out = False
            if only_version and v == only_version:
                filter_out = False
        except Exception:
            pass
    return filter_out


def _filter_os_version(connector, os_version_regex):
    if not os_version_regex:
        return False

    filter_out = True
    for instance in connector["status"].get("instances", []):
        try:
            if os_version_regex.match(_get_os_version(instance, "")):
                filter_out = False
        except Exception:
            pass
    return filter_out


def do_filter(
    ctx,
    no_down,
    connectors,
    not_version=None,
    only_version=None,
    sort_by=None,
    filter_os_version=None,
    **kwargs,
):
    if sort_by:
        connectors = sorted(connectors, key=lambda k: (operator.attrgetter(*sort_by)(k)))

    os_version_regex = None
    if filter_os_version:
        os_version_regex = re.compile(filter_os_version)
    filtered_connectors = []
    for connector in connectors:
        if _filter_version(connector, not_version, only_version):
            continue
        if _filter_os_version(connector, os_version_regex):
            continue
        if not no_down:
            filtered_connectors.append(connector)
        elif connector["status"]["operational_status"]["status"] != "down":
            filtered_connectors.append(connector)
    return filtered_connectors


def query(
    ctx,
    no_down=False,
    not_version=None,
    only_version=None,
    sort_by=None,
    page_at_id=None,
    page_size=500,
    filter_os_version=None,
    page_on=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    params = {}
    params["org_id"] = org_id
    if page_at_id is None:
        page_at_id = ""
    input_helpers.update_if_not_none(params, kwargs)
    params = pagination.normalize_page_args(params)
    if page_on is None:
        query_results = get_many_entries(
            apiclient.connectors_api.list_connector,
            "connectors",
            maximum=kwargs.get("limit", None),
            page_size=page_size,
            page_at_id=page_at_id,
            **params,
        )
    else:
        query_results = apiclient.connectors_api.list_connector(
            page_on=page_on, **params
        ).connectors
    return do_filter(
        ctx,
        no_down,
        query_results,
        only_version=only_version,
        not_version=not_version,
        sort_by=sort_by,
        filter_os_version=filter_os_version,
        **kwargs,
    )


def get(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    params = kwargs
    params["org_id"] = org_id
    input_helpers.update_if_not_none(params, kwargs)
    return apiclient.connectors_api.get_connector(connector_id, **params)


def list_connectors_simple_status(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    params = kwargs
    params["org_id"] = org_id
    input_helpers.update_if_not_none(params, kwargs)
    return apiclient.connectors_api.get_connectors_simple_status(**params).connectors


def get_instance(ctx, connector_id, connector_instance_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    params = kwargs
    params["org_id"] = org_id
    input_helpers.update_if_not_none(params, kwargs)
    return apiclient.connectors_api.get_instance(
        connector_id, connector_instance_id, **params
    )


def query_agents(ctx, column_format=None, filter_not_has_version=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    if column_format == "newformat":
        kwargs["show_stats"] = True

    if org_id:
        kwargs["org_id"] = org_id
    kwargs = pagination.normalize_page_args(kwargs)
    query_results = apiclient.connectors_api.list_agent_connector(
        **strip_none(kwargs)
    ).agent_connectors
    return query_results


def query_agent_instances(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    if org_id:
        kwargs["org_id"] = org_id
    return apiclient.connectors_api.list_instances(
        connector_id, **strip_none(kwargs)
    ).agent_connector_instances


def format_agent_instances(ctx, instances, **kwargs):
    columns = [
        metadata_column("id"),
        spec_column("org_id"),
        status_column("instance_number"),
        status_column("name"),
        status_column("service_account_id"),
        column("status", newname="status", getter=_get_oper_status, optional=True),
    ]

    return format_table(ctx, instances, columns)


def _instances_table():
    instance_columns = [
        status_column("instance_number"),
        column(
            "overall_status",
            newname="overall_status",
            getter=_get_oper_status(attrgetter("status.operational_status")),
            optional=True,
        ),
        column("status.stats.overall_status_info", newname="status_info", optional=True),
        column("hostname", getter=_get_hostname, optional=True),
        column("os_version", getter=_get_os_version, optional=True),
        column("version", getter=_get_version, optional=True),
        column(
            "stats",
            newname="last_status_change",
            getter=_get_oper_status_change,
            optional=True,
        ),
    ]
    return instance_columns


def format_connectors_as_text(
    ctx, connectors, sort_by=None, skip_sub_table=False, **kwargs
):
    instance_columns = _instances_table()
    columns = [
        metadata_column("id"),
        spec_column("name"),
        column(
            "stats",
            newname="last_status_change",
            getter=_get_oper_status_change,
            optional=True,
        ),
    ]
    if "metadata.created" in sort_by:
        columns.append(metadata_column("created"))

    if not get_org_from_input_or_ctx(ctx, **kwargs):
        columns.append(spec_column("org_id"))
        org_dictionary = get_org_by_dictionary(ctx, "")
        for connector in connectors:
            org = org_dictionary[0].get(connector.spec.org_id, {})
            connector.spec["organisation"] = org.get("organisation")
        columns.append(spec_column("organisation"))

    columns.append(
        column(
            "status",
            newname="status",
            getter=_get_oper_status(attrgetter("status.operational_status")),
            optional=True,
        )
    )
    if skip_sub_table is False:
        columns.append(
            subtable(ctx, "instances", instance_columns, subobject_name="status")
        )

    return format_table(ctx, connectors, columns)


windows_ver_regex = """^.*[\\s]+([\\d]+\\.[\\d]+\\.[\\d]+).*$"""
windows_ver = re.compile(windows_ver_regex)


def _truncate_os(os_ver: str):
    if not os_ver:
        return os_ver

    if os_ver.upper().startswith("MICROSOFT WINDOWS"):
        ver_num = windows_ver.findall(os_ver)
        if ver_num:
            return "Windows " + ver_num[0]

    return os_ver[0:20]


def _shorten(data, limit=20):
    if data:
        val = data[:limit] + (".." if len(data) > limit else "")
        return val.strip()


def _remove_utc(datetime_obj):
    return str(datetime_obj.replace(microsecond=0)).split("+")[0]


def _get_status(record):
    status = record.get("status")
    if not status:
        return
    return status.to_dict()


def _get_connection_uri(record, key):
    if "connection_uri" in record["spec"]:
        parsed_uri = urlparse(record["spec"]["connection_uri"])
        return parsed_uri.netloc
    return None


def _get_oper_status(operstatus_getter):
    def getit(record, key):
        status = operstatus_getter(record)["status"]
        if status == "down":
            return f"{Fore.LIGHTRED_EX}{status}{Fore.RESET}"
        elif status == "degraded":
            return f"{Fore.LIGHTYELLOW_EX}{status}{Fore.RESET}"
        elif status == "warn":
            return f"{Fore.LIGHTYELLOW_EX}{status}{Fore.RESET}"
        elif status == "good":
            return f"{Fore.LIGHTGREEN_EX}{status}{Fore.RESET}"
        return status

    return getit


def _get_hostname(record, key):
    return _shorten(_get_stats(record)["system"]["hostname"])


def _get_version(record, key):
    system = _get_stats(record)["system"]
    if "agent_version" in system:
        return _shorten(system["agent_version"], 16)
    elif "version" in system:
        return _shorten(system["version"], 16)


def _get_os_version(record, key):
    return _truncate_os(_get_stats(record)["system"]["os_version"])


def _get_oper_status_change(record, key):
    return _remove_utc(
        _get_status(record)["operational_status"].get("status_change_time")
    )


def _get_stats(record, key=None):
    status = record.get("status")
    if not status:
        return
    return status.to_dict()["stats"]


def format_agents_with_version(  # noqa
    ctx, agents, column_format=None, filter_not_has_version=None, **kwargs
):
    org_by_id = dict()
    try:
        org_by_id, _ = get_org_by_dictionary(ctx, "")
    except agilicus_api.exceptions.ForbiddenException:
        # Fall back on getting just our own
        org_by_id, _ = get_org_by_dictionary(ctx, None)

    def _get_hostname(record, key):
        return _shorten(_get_stats(record)["system"]["hostname"])

    def _get_agent_uptime(record, key):
        secs = _get_stats(record)["system"]["agent_uptime"]
        sec = timedelta(seconds=secs)
        d = datetime.datetime(1, 1, 1) + sec
        return f"{d.day - 1}:{d.hour}:{d.minute}:{d.second}"

    def _get_collection_time(record, key):
        return _remove_utc(_get_stats(record)["metadata"]["collection_time"])

    def _get_overall_status(record, key):
        return _get_stats(record)["overall_status"]

    def _get_org(record):
        spec = record.get("spec").to_dict()
        org_id = spec.get("org_id")
        return org_by_id.get(org_id, org_id)

    def _get_org_name(record, keys):
        return _get_org(record).get("organisation")

    def _get_org_contact(record, keys):
        return _get_org(record).get("contact_email")

    def _row_filter(record):
        if not filter_not_has_version:
            return True
        try:
            version = _get_version(record, None)
            if version != filter_not_has_version:
                return True
        except Exception:
            pass
        return False

    columns = [
        metadata_column("id"),
        spec_column("name"),
        column("stats", newname="version", getter=_get_version, optional=True),
        column("stats", newname="uptime", getter=_get_agent_uptime, optional=True),
        column("stats", newname="os_version", getter=_get_os_version, optional=True),
        column("stats", newname="last_seen", getter=_get_collection_time, optional=True),
        column("stats", newname="status", getter=_get_oper_status, optional=True),
        column(
            "stats",
            newname="last_status_change",
            getter=_get_oper_status_change,
            optional=True,
        ),
        column("stats", newname="hostname", getter=_get_hostname, optional=True),
        column("spec", newname="org", getter=_get_org_name, optional=True),
        column("spec", newname="contact", getter=_get_org_contact, optional=True),
    ]

    return format_table(
        ctx, agents, columns, getter=operator.itemgetter, row_filter=_row_filter
    )


def format_agents_as_text(ctx, agents, column_format=None, **kwargs):
    if column_format == "newformat":
        return format_agents_with_version(ctx, agents, **kwargs)

    app_service_columns = [
        column("id"),
        column("hostname"),
        column("port"),
        column("protocol"),
        column("service_type"),
    ]
    columns = [
        metadata_column("id"),
        spec_column("name"),
        spec_column("org_id"),
        spec_column("connection_uri"),
        spec_column("max_number_connections"),
        spec_column("local_authentication_enabled"),
        subtable(
            ctx, "application_services", app_service_columns, subobject_name="status"
        ),
    ]

    return format_table(ctx, agents, columns)


def add_agent(ctx, point_of_presence_tag=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)

    spec = agilicus.AgentConnectorSpec(org_id=org_id, **kwargs)
    if point_of_presence_tag:
        spec.connector_cloud_routing = agilicus.ConnectorCloudRouting(
            point_of_presence_tags=regions.tag_list_to_tag_names(point_of_presence_tag)
        )

    connector = agilicus.AgentConnector(spec=spec)
    return create_or_update(
        connector,
        lambda obj: apiclient.connectors_api.create_agent_connector(obj),
        lambda guid, obj: apiclient.connectors_api.replace_agent_connector(
            guid, agent_connector=obj
        ),
    )[0]


def get_agent(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    kwargs = strip_none(kwargs)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    if org_id:
        kwargs["org_id"] = org_id
    return apiclient.connectors_api.get_agent_connector(connector_id, **kwargs)


def delete_agent(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.delete_agent_connector(
        connector_id, org_id=org_id, **kwargs
    )


def get_agent_info(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.get_agent_info(connector_id, org_id=org_id, **kwargs)


def add_agent_local_bind(
    ctx,
    connector_id,
    bind_port,
    bind_host=None,
    revocation_proxy=False,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    connector = apiclient.connectors_api.get_agent_connector(connector_id, org_id=org_id)

    if revocation_proxy:
        routing = connector.spec.revocation_proxy
        if routing is None:
            routing = agilicus.AgentConnectorRouting(local_binds=[])
    else:
        routing = connector.spec.routing
        if routing is None:
            routing = agilicus.AgentConnectorRouting(local_binds=[])

    bind = agilicus.AgentConnectorLocalBind(bind_host=bind_host, bind_port=bind_port)
    routing.local_binds.append(bind)
    if revocation_proxy:
        connector.spec.revocation_proxy = routing
    else:
        connector.spec.routing = routing

    return _replace_agent(apiclient, connector_id=connector_id, connector=connector)


def delete_agent_local_bind(
    ctx,
    connector_id,
    bind_port=None,
    bind_host=None,
    revocation_proxy=False,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    connector = apiclient.connectors_api.get_agent_connector(connector_id, org_id=org_id)

    if revocation_proxy:
        routing = connector.spec.revocation_proxy
        if routing is None:
            return connector
    else:
        routing = connector.spec.routing
        if routing is None:
            return connector

    results = []
    for bind in routing.local_binds:
        if bind_port is not None and bind_port != bind.bind_port:
            results.append(bind)
            continue

        if bind_host is not None and bind_host != bind.bind_host:
            results.append(bind)
            continue

    routing.local_binds = results
    if revocation_proxy:
        connector.spec.revocation_proxy = routing
    else:
        connector.spec.routing = routing

    return _replace_agent(apiclient, connector_id=connector_id, connector=connector)


def add_agent_egress_gateway(
    ctx,
    connector_id,
    bind_port,
    bind_host=None,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    connector = apiclient.connectors_api.get_agent_connector(connector_id, org_id=org_id)

    gateway = connector.spec.egress_gateway
    if gateway is None:
        gateway = agilicus.EgressGateway(local_binds=[])

    bind = agilicus.AgentConnectorLocalBind(bind_host=bind_host, bind_port=bind_port)
    gateway.local_binds.append(bind)
    connector.spec.egress_gateway = gateway

    return _replace_agent(apiclient, connector_id=connector_id, connector=connector)


def delete_agent_egress_gateway(
    ctx,
    connector_id,
    bind_port=None,
    bind_host=None,
    **kwargs,
):
    apiclient = context.get_apiclient_from_ctx(ctx)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    connector = apiclient.connectors_api.get_agent_connector(connector_id, org_id=org_id)

    gateway = connector.spec.egress_gateway
    if gateway is None:
        return connector

    results = []
    for bind in gateway.local_binds:
        if bind_port is not None and bind_port != bind.bind_port:
            results.append(bind)
            continue

        if bind_host is not None and bind_host != bind.bind_host:
            results.append(bind)
            continue

    gateway.local_binds = results
    connector.spec.egress_gateway = gateway

    return _replace_agent(apiclient, connector_id=connector_id, connector=connector)


def _replace_agent(apiclient, connector_id, connector):

    # Clear out the status since it's unnecessary.
    del connector["status"]
    return apiclient.connectors_api.replace_agent_connector(
        connector_id, agent_connector=connector
    )


def replace_agent(
    ctx,
    connector_id,
    connection_uri=None,
    max_number_connections=None,
    name=None,
    service_account_required=None,
    local_authentication_enabled=None,
    name_slug=None,
    point_of_presence_tag=None,
    clear_point_of_presence_tags=False,
    proxy_tunnel_termination=None,
    dynamic_routes_enabled: Optional[Ternary] = None,
    on_demand_routes_enabled: Optional[Ternary] = None,
    admin_status=None,
    trap_disabled=None,
    revocation_proxy_trusted_cert_bundle_id=None,
    revocation_proxy_rules_bundle_id=None,
    ntp_forwarding_bind=None,
    ntp_forwarding_custom_bind=None,
    sync_local_clock=None,
    upstream_buffer_tuning=None,
    upstream_buffer_min_latency=None,
    upstream_buffer_max_latency=None,
    upstream_buffer_rmem_max=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)

    connector = apiclient.connectors_api.get_agent_connector(
        connector_id, org_id=org_id, **kwargs
    )

    if connection_uri:
        connector.spec.connection_uri = connection_uri

    if max_number_connections:
        connector.spec.max_number_connections = max_number_connections

    if name:
        connector.spec.name = name

    if service_account_required is not None:
        connector.spec.service_account_required = service_account_required

    if local_authentication_enabled is not None:
        connector.spec.local_authentication_enabled = local_authentication_enabled

    if name_slug is not None:
        connector.spec.name_slug = name_slug

    if clear_point_of_presence_tags:
        point_of_presence_tag = []

    if point_of_presence_tag is not None:
        tags = regions.tag_list_to_tag_names(point_of_presence_tag)
        cloud_routing = connector.spec.connector_cloud_routing
        if not cloud_routing:
            cloud_routing = agilicus.ConnectorCloudRouting(point_of_presence_tags=tags)
        else:
            cloud_routing.point_of_presence_tags = tags

        connector.spec.cloud_routing = cloud_routing

    _update_tunneling(
        connector, dynamic_routes_enabled, on_demand_routes_enabled, ntp_forwarding_bind
    )
    _update_ntp_forwarding(
        connector, ntp_forwarding_bind, ntp_forwarding_custom_bind, sync_local_clock
    )
    _update_upstream_buffer_control(
        connector,
        upstream_buffer_tuning,
        upstream_buffer_min_latency,
        upstream_buffer_max_latency,
        upstream_buffer_rmem_max,
    )

    if admin_status is not None:
        connector.spec.admin_status = admin_status

    if trap_disabled is not None:
        connector.spec.trap_disabled = trap_disabled

    if proxy_tunnel_termination is not None:
        connector.spec.proxy_tunnel_termination = proxy_tunnel_termination

    if revocation_proxy_trusted_cert_bundle_id is not None:
        connector.spec.revocation_proxy.trusted_cert_bundle = (
            revocation_proxy_trusted_cert_bundle_id
        )

    if revocation_proxy_rules_bundle_id is not None:
        connector.spec.revocation_proxy.rules_bundle = revocation_proxy_rules_bundle_id

    return _replace_agent(apiclient, connector_id=connector_id, connector=connector)


def _get_routing(
    connector: agilicus.AgentConnector,
) -> agilicus.AgentConnectorCloudRouting:
    if not connector.spec.routing:
        connector.spec.routing = agilicus.AgentConnectorCloudRouting(local_binds=[])
    return connector.spec.routing


def _get_buffer_control(connector) -> agilicus.UpstreamBufferControl:
    routing = _get_routing(connector)
    if not routing.upstream_buffer_control:
        routing.upstream_buffer_control = agilicus.UpstreamBufferControl()
    return routing.upstream_buffer_control


def _update_upstream_buffer_control(
    connector,
    upstream_buffer_tuning=None,
    upstream_buffer_min_latency=None,
    upstream_buffer_max_latency=None,
    upstream_buffer_rmem_max=None,
    **kwargs,
):
    if upstream_buffer_tuning is not None:
        _get_buffer_control(connector).upstream_buffer_tuning = upstream_buffer_tuning

    if upstream_buffer_min_latency is not None:
        _get_buffer_control(connector).min_latency = upstream_buffer_min_latency

    if upstream_buffer_max_latency is not None:
        _get_buffer_control(connector).max_latency = upstream_buffer_max_latency

    if upstream_buffer_rmem_max is not None:
        _get_buffer_control(connector).rmem_max = upstream_buffer_rmem_max


def _update_ntp_forwarding(
    connector, ntp_forwarding_bind, custom_bind, sync_local_clock
):
    routing = connector.spec.routing
    if sync_local_clock is not None:
        if routing is None:
            routing = agilicus.AgentConnectorCloudRouting(local_binds=[])
        connector.spec.routing = routing
        if sync_local_clock is not None:
            connector.spec.routing.sync_local_clock = sync_local_clock

    if ntp_forwarding_bind is None:
        return

    internal_networks = routing.internal_networks
    if not internal_networks:
        internal_networks = agilicus.InternalNetworkRouting()
    ntp = internal_networks.ntp
    if not ntp:
        ntp = agilicus.NTPInternalNetworkRouting()
    ntp.bind = agilicus.InternalNetworkBind(ntp_forwarding_bind)
    if custom_bind is not None:
        parts = urllib.parse.urlsplit("//" + custom_bind)
        ntp.bind.custom_bind = agilicus.AgentConnectorLocalBind(
            bind_host=parts.hostname, bind_port=parts.port or 123
        )
    internal_networks.ntp = ntp
    routing.internal_networks = internal_networks


def _update_tunneling(
    connector, dynamic_routes_enabled, on_demand_routes_enabled, ntp_forwarding_bind
):
    if dynamic_routes_enabled is None and on_demand_routes_enabled is None:
        return

    routing = connector.spec.routing
    if routing is None:
        routing = agilicus.AgentConnectorCloudRouting(local_binds=[])
    connector.spec.routing = routing
    tunneling = routing.tunneling
    if not tunneling:
        tunneling = agilicus.AgentConnectorTunneling()
        routing.tunneling = tunneling

    if dynamic_routes_enabled is not None:
        dynamic = dynamic_routes_enabled.to_bool_or_none()
        tunneling.dynamic_routes_enabled = dynamic

    if on_demand_routes_enabled is not None:
        on_demand = on_demand_routes_enabled.to_bool_or_none()
        tunneling.on_demand_routes_enabled = on_demand


def replace_agent_auth_info(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)

    info = agilicus.AgentLocalAuthInfo(**kwargs)
    return apiclient.connectors_api.replace_agent_connector_local_auth_info(
        connector_id, agent_local_auth_info=info
    )


def set_agent_connector_stats(ctx, connector_id, org_id, overall_status, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    system_objs = input_helpers.get_objects_by_location("system", kwargs)
    system = agilicus.AgentConnectorSystemStats(
        agent_connector_org_id=org_id, agent_connector_id=connector_id, **system_objs
    )
    transport_objs = input_helpers.get_objects_by_location("transport", kwargs)
    transport = agilicus.AgentConnectorTransportStats(**transport_objs)
    now = datetime.datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())
    metadata = agilicus.AgentConnectorStatsMetadata(collection_time=now)

    stats = agilicus.AgentConnectorStats(
        metadata=metadata,
        overall_status=overall_status,
        system=system,
        transport=transport,
    )

    return apiclient.connectors_api.create_agent_stats(connector_id, stats)


def get_agent_connector_stats(ctx, connector_id, org_id, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    return apiclient.connectors_api.get_agent_stats(
        connector_id, org_id=org_id, **kwargs
    )


def add_table_stats_rows(instancesStats, stat, value):
    if isinstance(value, dict):
        for nstat, nvalue in value.items():
            add_table_stats_rows(instancesStats, stat + ":" + nstat, nvalue)
    elif isinstance(value, list):
        for index, row in enumerate(value):
            try:
                if not isinstance(row, dict):
                    row = row.to_dict()
                for nstat, nvalue in row.items():
                    add_table_stats_rows(
                        instancesStats, stat + f"[{index}]:" + nstat, nvalue
                    )
            except Exception:
                pass
    else:
        instancesStats.setdefault(stat, []).append(value)


def show_agent_connector_stats(ctx, connector_id, **kwargs):
    headings = []
    headings.append("Statistic")
    results = query_agent_instances(ctx, connector_id, **kwargs)
    instance_stats = []
    for instance in results:
        instance_dict = instance.to_dict()
        stats = instance_dict.get("status", {}).get("stats", {})
        if not stats:
            continue

        for sub_key in ["sys_info", "perf_metrics"]:
            info = instance_dict.get("status", {}).get(sub_key, {})
            if info:
                stats.update({sub_key: info})

        instance_stats.append(stats)
        headings.append(instance.status.name)
    num_cols = len(results)

    table = PrettyTable(headings)

    instancesStats = {}
    for stats in instance_stats:
        for stat, value in stats.items():
            add_table_stats_rows(instancesStats, stat, value)
    for stat, entries in instancesStats.items():
        row = []
        row.append(stat)
        row.extend(entries[:num_cols] + [""] * (num_cols - len(entries)))
        table.add_row(row)
    table.align = "l"
    print(table)


def show_agent_connector_dynamic_stats(
    ctx, connector_id, org_id=None, detailed=False, breakdown=False, **kwargs
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    stats = get_agent_connector_dynamic_stats(ctx, connector_id, org_id, **kwargs)
    if not context.output_console(ctx):
        output.output_formatted(ctx, stats.to_dict())
        return

    headings = []
    headings.append("Statistic")
    headings.append("Value")
    table = PrettyTable(headings)
    table.add_row(("last_updated", stats.metadata.collection_time))
    stat_rows = _collate_upstream_stats(ctx, stats.upstream_totals, detailed)
    stat_rows.extend(_collate_forwarder_stats(ctx, stats.forwarder_totals, detailed))
    stat_rows.extend(_collate_audit_stats(ctx, stats.audit_totals, detailed))
    to_add_breakdown = []
    to_add_forwarder_breakdown = []
    to_add_audit_breakdown = []
    if breakdown:
        to_add_breakdown = stats.upstream_breakdown
        to_add_forwarder_breakdown = stats.forwarder_breakdown
        to_add_audit_breakdown = stats.audit_breakdown

    for item in to_add_breakdown:
        prefix = [item.connector_instance_id, item.application_service_id]
        stat_rows.extend(
            _collate_upstream_stats(ctx, item.upstream_stats, detailed, prefix)
        )

    for item in to_add_forwarder_breakdown:
        prefix = [item.connector_instance_id, item.forwarder_id]
        stat_rows.extend(
            _collate_forwarder_stats(ctx, item.forwarder_stats, detailed, prefix)
        )

    for item in to_add_audit_breakdown:
        prefix = [item.instance_id, item.audit_destination_id]
        stat_rows.extend(
            _collate_per_destination_stats(
                ctx, item.audit_destination_stats.to_dict(), detailed, prefix
            )
        )
    for stat, val in stat_rows:
        row = []
        row.append(stat)
        row.append(val)
        table.add_row(row)

    table.align = "l"
    print(table)


def list_connector_dynamic_stats(
    ctx, connector_ids, org_id=None, detailed=False, breakdown=False, **kwargs
):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    results = do_list_connector_stats(
        ctx, connector_ids, org_id, show_dynamic=True, **kwargs
    )
    if not context.output_console(ctx):
        output.output_formatted(ctx, results.to_dict()["dynamic_stats"])
        return

    headings = []
    headings.append("Statistic")
    headings.append("Value")
    table = PrettyTable(headings)
    for stats in results.dynamic_stats:
        base_prefix = [stats.connector_id]
        table.add_row((f"{base_prefix[0]}.last_updated", stats.metadata.collection_time))
        stat_rows = _collate_upstream_stats(
            ctx, stats.upstream_totals, detailed, base_prefix
        )
        to_add_breakdown = []
        if breakdown:
            to_add_breakdown = stats.upstream_breakdown

        for item in to_add_breakdown:
            prefix = base_prefix + [
                item.connector_instance_id,
                item.application_service_id,
            ]
            stat_rows.extend(
                _collate_upstream_stats(ctx, item.upstream_stats, detailed, prefix)
            )

        for stat, val in stat_rows:
            row = []
            row.append(stat)
            row.append(val)
            table.add_row(row)

    table.align = "l"
    print(table)


def list_connector_static_stats(ctx, connector_ids, org_id=None, **kwargs):
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    results = do_list_connector_stats(
        ctx, connector_ids, org_id, show_static=True, **kwargs
    )
    return results.static_stats


def format_static_stats_as_text(
    ctx,
    stats,
):
    instance_columns = _instances_table()
    columns = [
        column("connector_id"),
        column(
            "status",
            newname="status",
            getter=_get_oper_status(attrgetter("operational_status")),
            optional=True,
        ),
        subtable(ctx, "instances", instance_columns),
    ]

    return format_table(ctx, stats, columns)


def _collate_upstream_stats(ctx, upstream_stats, detailed, name_prefix=None):
    results = []
    upstream_stats = upstream_stats.to_dict()
    net_sum = upstream_stats.get("network_summary_stats")
    results.extend(_collate_stats_object(ctx, net_sum, name_prefix=name_prefix))
    http_sum = upstream_stats.get("http_summary_stats")
    results.extend(_collate_stats_object(ctx, http_sum, name_prefix=name_prefix))
    share_sum = upstream_stats.get("share_summary_stats")
    results.extend(_collate_stats_object(ctx, share_sum, name_prefix=name_prefix))

    if not detailed:
        return results

    net_sum = upstream_stats.get("network_detailed_stats")
    results.extend(_collate_stats_object(ctx, net_sum, name_prefix=name_prefix))
    http_sum = upstream_stats.get("http_detailed_stats")
    results.extend(_collate_stats_object(ctx, http_sum, name_prefix=name_prefix))
    share_sum = upstream_stats.get("share_detailed_stats")
    results.extend(_collate_stats_object(ctx, share_sum, name_prefix=name_prefix))
    return results


def _collate_forwarder_stats(ctx, forwarder_stats, detailed, name_prefix=None):
    results = []
    forwarder_stats = forwarder_stats.to_dict()
    summary = forwarder_stats.get("forwarder_summary_stats")

    def remap(name):
        return "forwarder_" + name

    results.extend(
        _collate_stats_object(ctx, summary, name_prefix=name_prefix, name_remap=remap)
    )

    if not detailed:
        return results

    detailed = forwarder_stats.get("forwarder_detailed_stats")
    results.extend(
        _collate_stats_object(ctx, detailed, name_prefix=name_prefix, name_remap=remap)
    )
    return results


def _collate_per_destination_stats(ctx, destination_stats, detailed, name_prefix=None):
    results = []
    summary = destination_stats.get("destination_summary_stats")

    def remap(name):
        return "audit_" + name

    results.extend(
        _collate_stats_object(ctx, summary, name_prefix=name_prefix, name_remap=remap)
    )

    if not detailed:
        return results

    detailed = destination_stats.get("destination_detailed_stats")
    results.extend(
        _collate_stats_object(ctx, detailed, name_prefix=name_prefix, name_remap=remap)
    )

    return results


def _collate_audit_stats(ctx, audit_stats, detailed, name_prefix=None):
    results = []
    audit_stats = audit_stats.to_dict()

    def remap(name):
        return "audit_" + name

    system = audit_stats.get("system_stats")
    results.extend(
        _collate_stats_object(ctx, system, name_prefix=name_prefix, name_remap=remap)
    )

    destination_stats = audit_stats.get("destination_stats")
    results.extend(
        _collate_per_destination_stats(ctx, destination_stats, detailed, name_prefix)
    )
    return results


def _collate_stats_object(ctx, stats_obj, name_prefix=None, name_remap=lambda x: x):
    if not stats_obj:
        return []
    results = []
    if name_prefix is None:
        name_prefix = []
    for key, val in stats_obj.items():
        name = name_prefix + [name_remap(key)]
        if isinstance(val, dict):
            results.extend(_collate_stats_object(ctx, val, name))
            continue
        results.append((".".join(name), val))
    return results


def get_agent_connector_dynamic_stats(
    ctx, connector_id, org_id, collected_since: datetime.datetime, **kwargs
) -> agilicus_api.AgentConnectorDynamicStats:
    apiclient = context.get_apiclient_from_ctx(ctx)

    if collected_since is not None:
        collected_since = collected_since.replace(tzinfo=datetime.timezone.utc)
        kwargs["collected_since"] = collected_since
    results = apiclient.connectors_api.get_agent_connector_dynamic_stats(
        connector_id=connector_id, org_id=org_id, **kwargs
    )
    return results


def do_list_connector_stats(
    ctx,
    connector_ids,
    org_id,
    collected_since: Optional[datetime.datetime] = None,
    show_dynamic=False,
    show_static=False,
    **kwargs,
) -> agilicus_api.AgentConnectorDynamicStats:
    apiclient = context.get_apiclient_from_ctx(ctx)

    if collected_since is not None:
        collected_since = collected_since.replace(tzinfo=datetime.timezone.utc)
        kwargs["collected_since"] = collected_since

    results = apiclient.connectors_api.list_connector_stats(
        connector_id_list=connector_ids,
        org_id=org_id,
        show_dynamic=show_dynamic,
        show_static=show_static,
        **kwargs,
    )
    return results


def query_ipsec(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    kwargs["org_id"] = org_id
    kwargs = strip_none(kwargs)
    query_results = apiclient.connectors_api.list_ipsec_connector(**kwargs)
    return query_results


def add_ipsec(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)

    spec = agilicus.IpsecConnectorSpec(org_id=org_id, **kwargs)
    connector = agilicus.IpsecConnector(spec=spec)
    return apiclient.connectors_api.create_ipsec_connector(connector)


def add_or_update_ipsec_connection(
    ctx,
    connector_id,
    name,
    org_id=None,
    inherit_from=None,
    remote_ipv4_block=None,
    ike_chain_of_trust_certificates_filename=None,
    update_connection=False,
    **kwargs,
):

    kwargs = strip_none(kwargs)
    if ike_chain_of_trust_certificates_filename is not None:
        ike_chain_of_trust_certificates = open(
            ike_chain_of_trust_certificates_filename, "r"
        ).read()
        kwargs["ike_chain_of_trust_certificates"] = ike_chain_of_trust_certificates

    if remote_ipv4_block:
        remote_ipv4_ranges = []
        for block in remote_ipv4_block:
            remote_ipv4_ranges.append(
                agilicus.IpsecConnectionIpv4Block(ipv4_address_block=block)
            )
        kwargs["remote_ipv4_ranges"] = remote_ipv4_ranges

    connector = get_ipsec(ctx, connector_id, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    if update_connection:
        new_connections = []
        for _connection in connector.spec.connections:
            if _connection.name == name:
                # update the connection with the updated kwargs by just blasting the new
                # values in to it.
                for k, v in kwargs.items():
                    _connection.spec[k] = v
                connection = _connection
                if inherit_from is not None:
                    connection.inherit_from = inherit_from

            new_connections.append(_connection)
        connector.spec.connections = new_connections
    else:
        connection_spec = input_helpers.model_from_dict(
            agilicus.IpsecConnectionSpec, kwargs
        )
        connection = agilicus.IpsecConnection(name, spec=connection_spec)
        if inherit_from:
            connection.inherit_from = inherit_from
        connector.spec.connections.append(connection)

    return apiclient.connectors_api.replace_ipsec_connector(
        connector_id, ipsec_connector=connector
    )


def delete_ipsec_connection(ctx, connector_id, name, org_id=None, **kwargs):
    connector = get_ipsec(ctx, connector_id, org_id=org_id)
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    kwargs = strip_none(kwargs)
    update_connections = []
    for connection in connector.spec.connections:
        if connection.name != name:
            update_connections.append(connection)
    connector.spec.connections = update_connections

    return apiclient.connectors_api.replace_ipsec_connector(
        connector_id, ipsec_connector=connector
    )


def get_ipsec(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.get_ipsec_connector(
        connector_id, org_id=org_id, **kwargs
    )


def delete_ipsec(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.delete_ipsec_connector(
        connector_id, org_id=org_id, **kwargs
    )


def get_ipsec_info(ctx, connector_id, org_id=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.get_ipsec_connector_info(
        connector_id, org_id=org_id, **kwargs
    )


def replace_ipsec(
    ctx,
    connector_id,
    name=None,
    name_slug=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)

    connector = apiclient.connectors_api.get_ipsec_connector(
        connector_id, org_id=org_id, **kwargs
    )

    if name:
        connector.spec.name = name

    if name_slug:
        connector.spec.name_slug = name_slug

    return apiclient.connectors_api.replace_ipsec_connector(
        connector_id, ipsec_connector=connector
    )


def show_connectors_usage_metrics(ctx, org_ids, **kwargs):
    apiclient = context.get_apiclient_from_ctx(ctx)
    if not org_ids:
        raise Exception("require at least one org_id")
    return apiclient.connectors_api.get_connector_usage_metrics(org_ids=org_ids)


def format_queues(ctx, queues):
    columns = [
        metadata_column("id"),
        spec_column("connector_id"),
        spec_column("instance_name"),
        spec_column("org_id"),
        spec_column("queue_ttl"),
        status_column("queue_name"),
        status_column("expired"),
    ]

    return format_table(ctx, queues.queues, columns)


def get_connector_queues(ctx, connector_id=None, org_id=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    params = kwargs
    if org_id is None:
        params["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    input_helpers.update_if_not_none(params, kwargs)
    if connector_id is not None:
        return apiclient.connectors_api.get_connector_queues(connector_id, **params)
    else:
        return apiclient.connectors_api.get_queues(**params)


def add_connector_queue(
    ctx,
    connector_id=None,
    instance_name=None,
    queue_ttl=None,
    dynamic_routes_enabled=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    spec = agilicus_api.AgentConnectorQueueSpec(
        connector_id=connector_id, instance_name=instance_name, org_id=org_id
    )

    if queue_ttl is not None:
        spec.queue_ttl = queue_ttl

    if dynamic_routes_enabled is not None:
        spec.dynamic_routes_enabled = True
    queue = agilicus_api.AgentConnectorQueue(spec=spec)

    return apiclient.connectors_api.create_queue(connector_id, queue)


def delete_connector_queue(ctx, connector_id, queue_id, org_id=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    params = kwargs
    if org_id is None:
        params["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    input_helpers.update_if_not_none(params, kwargs)
    return apiclient.connectors_api.delete_connector_queue(
        connector_id, queue_id, **params
    )


def delete_agent_connector_instance(ctx, connector_id, connector_instance_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.delete_instance(
        connector_id,
        connector_instance_id=connector_instance_id,
        org_id=org_id,
    )


def get_connector_stats_config(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    kwargs["org_id"] = get_org_from_input_or_ctx(ctx, **kwargs)
    return apiclient.connectors_api.get_stats_config(connector_id, **kwargs)


def configure_stats_publishing(
    ctx,
    org_id,
    connector_id,
    publish_period_s,
    net_summary_duration_s,
    http_summary_duration_s,
    net_detailed_duration_s,
    http_detailed_duration_s,
    share_summary_duration_s,
    share_detailed_duration_s,
    forwarder_summary_duration_s,
    forwarder_detailed_duration_s,
    audit_summary_duration_s,
    audit_detailed_duration_s,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, org_id=org_id)

    net = agilicus_api.StatsPublishingLevelConfig()
    http = agilicus_api.StatsPublishingLevelConfig()
    share = agilicus_api.StatsPublishingLevelConfig()
    forwarder = agilicus_api.StatsPublishingLevelConfig()
    audit = agilicus_api.StatsPublishingLevelConfig()

    if net_summary_duration_s:
        net.summary_duration_seconds = net_summary_duration_s
    if net_detailed_duration_s:
        net.detailed_duration_seconds = net_detailed_duration_s
    if http_summary_duration_s:
        http.summary_duration_seconds = http_summary_duration_s
    if http_detailed_duration_s:
        http.detailed_duration_seconds = http_detailed_duration_s
    if share_summary_duration_s:
        share.summary_duration_seconds = share_summary_duration_s
    if share_detailed_duration_s:
        share.detailed_duration_seconds = share_detailed_duration_s
    if forwarder_summary_duration_s:
        forwarder.summary_duration_seconds = forwarder_summary_duration_s
    if forwarder_detailed_duration_s:
        forwarder.detailed_duration_seconds = forwarder_detailed_duration_s

    if audit_summary_duration_s:
        audit.summary_duration_seconds = audit_summary_duration_s
    if audit_detailed_duration_s:
        audit.detailed_duration_seconds = audit_detailed_duration_s
    cfg = agilicus_api.StatsPublishingConfig(
        upstream_network_publishing=net,
        upstream_http_publishing=http,
        upstream_share_publishing=share,
        forwarder_publishing=forwarder,
        audit_publishing=audit,
        publish_period_seconds=publish_period_s,
    )
    req = agilicus_api.ConfigureConnectorStatsPublishingRequest(
        org_id, connector_ids=list(connector_id), stats_publishing_config=cfg
    )
    return apiclient.connectors_api.create_configure_publishing_request(req)


def format_for_gc(ctx, connectors):
    columns = [
        metadata_column("id"),
        spec_column("org_id", "org id"),
        spec_column("name"),
    ]
    return format_table(ctx, connectors, columns)


def list_connector_proxies(
    ctx,
    page_at_id=None,
    page_size=500,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    params = {}
    params["org_id"] = org_id
    if page_at_id is None:
        page_at_id = ""
    input_helpers.update_if_not_none(params, kwargs)
    return get_many_entries(
        apiclient.connectors_api.list_proxies,
        "connector_proxies",
        maximum=kwargs.get("limit", None),
        page_size=page_size,
        page_at_id=page_at_id,
        **params,
    )


def add_connector_proxy(ctx, bind_host=None, bind_port=None, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs = strip_none(kwargs)

    spec = agilicus.AgentConnectorProxySpec(org_id=org_id, **kwargs)

    if bind_port or bind_host:
        local_bind = agilicus_api.AgentConnectorLocalBind(
            bind_host=bind_host,
            bind_port=bind_port,
        )
        spec.local_bind = local_bind

    return apiclient.connectors_api.create_connector_proxy(
        agilicus.AgentConnectorProxy(spec),
    )


def delete_connector_proxy(ctx, id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs = strip_none(kwargs)

    return apiclient.connectors_api.delete_proxy(
        id,
        org_id=org_id,
    )


def query_services(ctx, connector_id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    kwargs["org_id"] = org_id
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.list_services(connector_id, **kwargs).services


def add_service(ctx, connector_id, service, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    spec = agilicus.ConnectorServiceSpec(
        org_id=org_id,
        service=service,
        connector_id=connector_id,
    )
    connectorService = agilicus.ConnectorService(spec=spec)
    return apiclient.connectors_api.create_service(connector_id, connectorService)


def delete_service(ctx, connector_id, service, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs["org_id"] = org_id
    kwargs = strip_none(kwargs)
    return apiclient.connectors_api.delete_service(
        connector_id, connector_service=service, **kwargs
    )



import agilicus
from . import apps
from . import context
from .input_helpers import get_org_from_input_or_ctx
from .output.table import (
    spec_column,
    format_table,
    metadata_column,
)


def add(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    spec = agilicus.ServiceForwarderSpec(org_id=org_id, **kwargs)
    forwarder = agilicus.ServiceForwarder(spec=spec)
    return apiclient.app_services_api.create_service_forwarder(forwarder)


def query(ctx, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)
    org_id = get_org_from_input_or_ctx(ctx, **kwargs)

    kwargs["org_id"] = org_id
    query_results = apiclient.app_services_api.list_service_forwarders(**kwargs)
    return query_results.service_forwarders


def get(ctx, id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    return apiclient.app_services_api.get_service_forwarder(id, org_id=org_id, **kwargs)


def delete(ctx, id, **kwargs):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)
    return apiclient.app_services_api.delete_service_forwarder(
        id, org_id=org_id, **kwargs
    )


def replace(
    ctx,
    id,
    name=None,
    port=None,
    connector_id=None,
    application_service_id=None,
    bind_address=None,
    protocol=None,
    port_range=None,
    source_port_override=None,
    **kwargs,
):
    token = context.get_token(ctx)
    apiclient = context.get_apiclient(ctx, token)

    org_id = get_org_from_input_or_ctx(ctx, **kwargs)
    kwargs.pop("org_id", None)

    forwarder = apiclient.app_services_api.get_service_forwarder(
        id, org_id=org_id, **kwargs
    )

    if name:
        forwarder.spec.name = name

    if port:
        forwarder.spec.port = port

    if connector_id:
        forwarder.spec.connector_id = connector_id

    if application_service_id:
        forwarder.spec.application_service_id = application_service_id

    if bind_address:
        forwarder.spec.bind_address = bind_address

    if protocol:
        forwarder.spec.protocol = protocol

    if port_range is not None:
        forwarder.spec.config = apps.configure_port(forwarder.spec.config, port_range)

    if source_port_override is not None:
        if forwarder.spec.config is None:
            forwarder.spec.config = agilicus.NetworkServiceConfig()
        forwarder.spec.config.source_port_override = apps.parse_ports(
            source_port_override
        )
    return apiclient.app_services_api.replace_service_forwarder(
        id, service_forwarder=forwarder
    )


def format_forwarders_as_text(ctx, forwarders):
    columns = [
        metadata_column("id"),
        spec_column("name"),
        spec_column("org_id"),
        spec_column("bind_address"),
        spec_column("port"),
        spec_column("protocol"),
        spec_column("connector_id"),
        spec_column("application_service_id", "app_service_id"),
    ]
    return format_table(ctx, forwarders, columns)



from unittest.mock import MagicMock

class PolicyTemplatesApiMock:

    def __init__(self):
        self.mock_create_policy_template_instance = MagicMock()
        self.mock_delete_policy_template_instance = MagicMock()
        self.mock_get_policy_template_instance = MagicMock()
        self.mock_list_policy_template_instances = MagicMock()
        self.mock_replace_policy_template_instance = MagicMock()

    def create_policy_template_instance(self, *args, **kwargs):
        """
        This method mocks the original api PolicyTemplatesApi.create_policy_template_instance with MagicMock.
        """
        return self.mock_create_policy_template_instance(self, *args, **kwargs)

    def delete_policy_template_instance(self, *args, **kwargs):
        """
        This method mocks the original api PolicyTemplatesApi.delete_policy_template_instance with MagicMock.
        """
        return self.mock_delete_policy_template_instance(self, *args, **kwargs)

    def get_policy_template_instance(self, *args, **kwargs):
        """
        This method mocks the original api PolicyTemplatesApi.get_policy_template_instance with MagicMock.
        """
        return self.mock_get_policy_template_instance(self, *args, **kwargs)

    def list_policy_template_instances(self, *args, **kwargs):
        """
        This method mocks the original api PolicyTemplatesApi.list_policy_template_instances with MagicMock.
        """
        return self.mock_list_policy_template_instances(self, *args, **kwargs)

    def replace_policy_template_instance(self, *args, **kwargs):
        """
        This method mocks the original api PolicyTemplatesApi.replace_policy_template_instance with MagicMock.
        """
        return self.mock_replace_policy_template_instance(self, *args, **kwargs)




from agilicus.agilicus_api import Oauth2Auth


def get_oauth2_auth(properties, pop=False):
    existing = properties.get("oauth2", {})
    # Find all oauth2_-prefixed items
    matched = []
    for key, value in properties.items():
        parts = key.split("oauth2_", 2)
        if len(parts) != 2:
            continue
        matched.append(key)
        if value is None:
            continue
        if isinstance(value, tuple):
            value = list(value)
        existing[parts[1]] = value

    if matched and pop:
        for key in matched:
            properties.pop(key)

    if not existing:
        return None

    return Oauth2Auth(**existing)



"""
    Agilicus API

    Agilicus is API-first. Modern software is controlled by other software, is open, is available for you to use the way you want, securely, simply.  The OpenAPI Specification in YAML format is available on [www](https://www.agilicus.com/www/api/agilicus-openapi.yaml) for importing to other tools.  A rendered, online viewable and usable version of this specification is available at [api](https://www.agilicus.com/api). You may try the API inline directly in the web page. To do so, first obtain an Authentication Token (the simplest way is to install the Python SDK, and then run `agilicus-cli --issuer https://MYISSUER get-token`). You will need an org-id for most calls (and can obtain from `agilicus-cli --issuer https://MYISSUER list-orgs`). The `MYISSUER` will typically be `auth.MYDOMAIN`, and you will see it as you sign-in to the administrative UI.  This API releases on Bearer-Token authentication. To obtain a valid bearer token you will need to Authenticate to an Issuer with OpenID Connect (a superset of OAUTH2).  Your \"issuer\" will look like https://auth.MYDOMAIN. For example, when you signed-up, if you said \"use my own domain name\" and assigned a CNAME of cloud.example.com, then your issuer would be https://auth.cloud.example.com.  If you selected \"use an Agilicus supplied domain name\", your issuer would look like https://auth.myorg.agilicus.cloud.  For test purposes you can use our [Python SDK](https://pypi.org/project/agilicus/) and run `agilicus-cli --issuer https://auth.MYDOMAIN get-token`.  This API may be used in any language runtime that supports OpenAPI 3.0, or, you may use our [Python SDK](https://pypi.org/project/agilicus/), our [Typescript SDK](https://www.npmjs.com/package/@agilicus/angular), or our [Golang SDK](https://git.agilicus.com/pub/sdk-go).  100% of the activities in our system our API-driven, from our web-admin, through our progressive web applications, to all internals: there is nothing that is not accessible.  For more information, see [developer resources](https://www.agilicus.com/developer).   # noqa: E501

    The version of the OpenAPI document: 2026.03.25
    Contact: dev@agilicus.com
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from agilicus_api.api_client import ApiClient, Endpoint as _Endpoint
from agilicus_api.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types
)
from agilicus_api.model.deployment import Deployment
from agilicus_api.model.deployment_instance import DeploymentInstance
from agilicus_api.model.deployment_template import DeploymentTemplate
from agilicus_api.model.error_message import ErrorMessage
from agilicus_api.model.list_deployment_instances import ListDeploymentInstances
from agilicus_api.model.list_deployment_templates import ListDeploymentTemplates
from agilicus_api.model.list_deployments import ListDeployments


class DeploymentsApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        def __create_deployment(
            self,
            deployment,
            **kwargs
        ):
            """Add a Deployment  # noqa: E501

            Create a new Deployment   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_deployment(deployment, async_req=True)
            >>> result = thread.get()

            Args:
                deployment (Deployment):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Deployment
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment'] = \
                deployment
            return self.call_with_http_info(**kwargs)

        if self.create_deployment is None:
            self.create_deployment = _Endpoint(
                settings={
                    'response_type': (Deployment,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployments',
                    'operation_id': 'create_deployment',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment',
                    ],
                    'required': [
                        'deployment',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment':
                            (Deployment,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'deployment': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_deployment
            )

        def __create_deployment_instance(
            self,
            deployment_instance,
            **kwargs
        ):
            """Add a DeploymentInstance  # noqa: E501

            Create a new DeploymentInstance   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_deployment_instance(deployment_instance, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_instance (DeploymentInstance):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentInstance
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_instance'] = \
                deployment_instance
            return self.call_with_http_info(**kwargs)

        if self.create_deployment_instance is None:
            self.create_deployment_instance = _Endpoint(
                settings={
                    'response_type': (DeploymentInstance,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_instances',
                    'operation_id': 'create_deployment_instance',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_instance',
                    ],
                    'required': [
                        'deployment_instance',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_instance':
                            (DeploymentInstance,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'deployment_instance': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_deployment_instance
            )

        def __create_deployment_template(
            self,
            deployment_template,
            **kwargs
        ):
            """Add a DeploymentTemplate  # noqa: E501

            Create a new DeploymentTemplate   # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.create_deployment_template(deployment_template, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_template (DeploymentTemplate):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentTemplate
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_template'] = \
                deployment_template
            return self.call_with_http_info(**kwargs)

        if self.create_deployment_template is None:
            self.create_deployment_template = _Endpoint(
                settings={
                    'response_type': (DeploymentTemplate,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_templates',
                    'operation_id': 'create_deployment_template',
                    'http_method': 'POST',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_template',
                    ],
                    'required': [
                        'deployment_template',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_template':
                            (DeploymentTemplate,),
                    },
                    'attribute_map': {
                    },
                    'location_map': {
                        'deployment_template': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__create_deployment_template
            )

        def __delete_deployment(
            self,
            deployment_id,
            **kwargs
        ):
            """Delete a deployment  # noqa: E501

            Delete a Deployment  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_deployment(deployment_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_id (str): deployment id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_id'] = \
                deployment_id
            return self.call_with_http_info(**kwargs)

        if self.delete_deployment is None:
            self.delete_deployment = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployments/{deployment_id}',
                    'operation_id': 'delete_deployment',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_id': 'deployment_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_deployment
            )

        def __delete_deployment_instance(
            self,
            deployment_instance_id,
            **kwargs
        ):
            """Delete a DeploymentInstance  # noqa: E501

            Delete a DeploymentInstance  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_deployment_instance(deployment_instance_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_instance_id (str): deployment instance id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_instance_id'] = \
                deployment_instance_id
            return self.call_with_http_info(**kwargs)

        if self.delete_deployment_instance is None:
            self.delete_deployment_instance = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_instances/{deployment_instance_id}',
                    'operation_id': 'delete_deployment_instance',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_instance_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_instance_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_instance_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_instance_id': 'deployment_instance_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_instance_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_deployment_instance
            )

        def __delete_deployment_template(
            self,
            deployment_template_id,
            **kwargs
        ):
            """Delete a DeploymentTemplate  # noqa: E501

            Delete a DeploymentTemplate  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.delete_deployment_template(deployment_template_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_template_id (str): deployment template id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                None
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_template_id'] = \
                deployment_template_id
            return self.call_with_http_info(**kwargs)

        if self.delete_deployment_template is None:
            self.delete_deployment_template = _Endpoint(
                settings={
                    'response_type': None,
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_templates/{deployment_template_id}',
                    'operation_id': 'delete_deployment_template',
                    'http_method': 'DELETE',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_template_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_template_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_template_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_template_id': 'deployment_template_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_template_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__delete_deployment_template
            )

        def __get_deployment(
            self,
            deployment_id,
            **kwargs
        ):
            """Get a Deployment  # noqa: E501

            Get a Deployment  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_deployment(deployment_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_id (str): deployment id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Deployment
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_id'] = \
                deployment_id
            return self.call_with_http_info(**kwargs)

        if self.get_deployment is None:
            self.get_deployment = _Endpoint(
                settings={
                    'response_type': (Deployment,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployments/{deployment_id}',
                    'operation_id': 'get_deployment',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_id': 'deployment_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_deployment
            )

        def __get_deployment_instance(
            self,
            deployment_instance_id,
            **kwargs
        ):
            """Get a DeploymentInstance  # noqa: E501

            Get a DeploymentInstance  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_deployment_instance(deployment_instance_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_instance_id (str): deployment instance id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentInstance
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_instance_id'] = \
                deployment_instance_id
            return self.call_with_http_info(**kwargs)

        if self.get_deployment_instance is None:
            self.get_deployment_instance = _Endpoint(
                settings={
                    'response_type': (DeploymentInstance,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_instances/{deployment_instance_id}',
                    'operation_id': 'get_deployment_instance',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_instance_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_instance_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_instance_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_instance_id': 'deployment_instance_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_instance_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_deployment_instance
            )

        def __get_deployment_template(
            self,
            deployment_template_id,
            **kwargs
        ):
            """Get a DeploymentTemplate  # noqa: E501

            Get a DeploymentTemplate  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.get_deployment_template(deployment_template_id, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_template_id (str): deployment template id path

            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentTemplate
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_template_id'] = \
                deployment_template_id
            return self.call_with_http_info(**kwargs)

        if self.get_deployment_template is None:
            self.get_deployment_template = _Endpoint(
                settings={
                    'response_type': (DeploymentTemplate,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_templates/{deployment_template_id}',
                    'operation_id': 'get_deployment_template',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_template_id',
                        'org_id',
                    ],
                    'required': [
                        'deployment_template_id',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_template_id':
                            (str,),
                        'org_id':
                            (str,),
                    },
                    'attribute_map': {
                        'deployment_template_id': 'deployment_template_id',
                        'org_id': 'org_id',
                    },
                    'location_map': {
                        'deployment_template_id': 'path',
                        'org_id': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__get_deployment_template
            )

        def __list_deployment_instances(
            self,
            **kwargs
        ):
            """Get all deployment instances  # noqa: E501

            Get a list DeploymentInstance objects  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_deployment_instances(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListDeploymentInstances
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_deployment_instances is None:
            self.list_deployment_instances = _Endpoint(
                settings={
                    'response_type': (ListDeploymentInstances,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_instances',
                    'operation_id': 'list_deployment_instances',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                        'limit',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                        'limit':
                            (int,),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                        'limit': 'limit',
                    },
                    'location_map': {
                        'org_id': 'query',
                        'limit': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_deployment_instances
            )

        def __list_deployment_templates(
            self,
            **kwargs
        ):
            """Get all deployment templates  # noqa: E501

            Get a list DeploymentTemplate objects  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_deployment_templates(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                deployment_template_type (str): deployment template type query. [optional]
                system_templates (bool): include system templates with the query. [optional]
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListDeploymentTemplates
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_deployment_templates is None:
            self.list_deployment_templates = _Endpoint(
                settings={
                    'response_type': (ListDeploymentTemplates,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_templates',
                    'operation_id': 'list_deployment_templates',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                        'limit',
                        'deployment_template_type',
                        'system_templates',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                        'deployment_template_type',
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                        ('deployment_template_type',): {

                            "MODEL": "model",
                            "SCHEMA": "schema"
                        },
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                        'limit':
                            (int,),
                        'deployment_template_type':
                            (str,),
                        'system_templates':
                            (bool,),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                        'limit': 'limit',
                        'deployment_template_type': 'deployment_template_type',
                        'system_templates': 'system_templates',
                    },
                    'location_map': {
                        'org_id': 'query',
                        'limit': 'query',
                        'deployment_template_type': 'query',
                        'system_templates': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_deployment_templates
            )

        def __list_deployments(
            self,
            **kwargs
        ):
            """Get all deployments  # noqa: E501

            Get a list Deployment objects  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.list_deployments(async_req=True)
            >>> result = thread.get()


            Keyword Args:
                org_id (str): Organisation Unique identifier. [optional]
                limit (int): limit the number of rows in the response. [optional] if omitted the server will use the default value of 500
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                ListDeployments
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            return self.call_with_http_info(**kwargs)

        if self.list_deployments is None:
            self.list_deployments = _Endpoint(
                settings={
                    'response_type': (ListDeployments,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployments',
                    'operation_id': 'list_deployments',
                    'http_method': 'GET',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'org_id',
                        'limit',
                    ],
                    'required': [],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                        'limit',
                    ]
                },
                root_map={
                    'validations': {
                        ('limit',): {

                            'inclusive_maximum': 500,
                            'inclusive_minimum': 1,
                        },
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'org_id':
                            (str,),
                        'limit':
                            (int,),
                    },
                    'attribute_map': {
                        'org_id': 'org_id',
                        'limit': 'limit',
                    },
                    'location_map': {
                        'org_id': 'query',
                        'limit': 'query',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [],
                },
                api_client=api_client,
                callable=__list_deployments
            )

        def __update_deployment(
            self,
            deployment_id,
            deployment,
            **kwargs
        ):
            """Update a Deployment  # noqa: E501

            Update a Deployment  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.update_deployment(deployment_id, deployment, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_id (str): deployment id path
                deployment (Deployment):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                Deployment
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_id'] = \
                deployment_id
            kwargs['deployment'] = \
                deployment
            return self.call_with_http_info(**kwargs)

        if self.update_deployment is None:
            self.update_deployment = _Endpoint(
                settings={
                    'response_type': (Deployment,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployments/{deployment_id}',
                    'operation_id': 'update_deployment',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_id',
                        'deployment',
                    ],
                    'required': [
                        'deployment_id',
                        'deployment',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_id':
                            (str,),
                        'deployment':
                            (Deployment,),
                    },
                    'attribute_map': {
                        'deployment_id': 'deployment_id',
                    },
                    'location_map': {
                        'deployment_id': 'path',
                        'deployment': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__update_deployment
            )

        def __update_deployment_instance(
            self,
            deployment_instance_id,
            deployment_instance,
            **kwargs
        ):
            """Update a DeploymentInstance  # noqa: E501

            Update a DeploymentInstance  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.update_deployment_instance(deployment_instance_id, deployment_instance, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_instance_id (str): deployment instance id path
                deployment_instance (DeploymentInstance):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentInstance
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_instance_id'] = \
                deployment_instance_id
            kwargs['deployment_instance'] = \
                deployment_instance
            return self.call_with_http_info(**kwargs)

        if self.update_deployment_instance is None:
            self.update_deployment_instance = _Endpoint(
                settings={
                    'response_type': (DeploymentInstance,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_instances/{deployment_instance_id}',
                    'operation_id': 'update_deployment_instance',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_instance_id',
                        'deployment_instance',
                    ],
                    'required': [
                        'deployment_instance_id',
                        'deployment_instance',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_instance_id':
                            (str,),
                        'deployment_instance':
                            (DeploymentInstance,),
                    },
                    'attribute_map': {
                        'deployment_instance_id': 'deployment_instance_id',
                    },
                    'location_map': {
                        'deployment_instance_id': 'path',
                        'deployment_instance': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__update_deployment_instance
            )

        def __update_deployment_template(
            self,
            deployment_template_id,
            deployment_template,
            **kwargs
        ):
            """Update a DeploymentTemplate  # noqa: E501

            Update a DeploymentTemplate  # noqa: E501
            This method makes a synchronous HTTP request by default. To make an
            asynchronous HTTP request, please pass async_req=True

            >>> thread = api.update_deployment_template(deployment_template_id, deployment_template, async_req=True)
            >>> result = thread.get()

            Args:
                deployment_template_id (str): deployment template id path
                deployment_template (DeploymentTemplate):

            Keyword Args:
                _return_http_data_only (bool): response data without head status
                    code and headers. Default is True.
                _preload_content (bool): if False, the urllib3.HTTPResponse object
                    will be returned without reading/decoding response data.
                    Default is True.
                _request_timeout (float/tuple): timeout setting for this request. If one
                    number provided, it will be total request timeout. It can also
                    be a pair (tuple) of (connection, read) timeouts.
                    Default is None.
                _check_input_type (bool): specifies if type checking
                    should be done one the data sent to the server.
                    Default is True.
                _check_return_type (bool): specifies if type checking
                    should be done one the data received from the server.
                    Default is True.
                _host_index (int/None): specifies the index of the server
                    that we want to use.
                    Default is read from the configuration.
                async_req (bool): execute request asynchronously

            Returns:
                DeploymentTemplate
                    If the method is called asynchronously, returns the request
                    thread.
            """
            kwargs['async_req'] = kwargs.get(
                'async_req', False
            )
            kwargs['_return_http_data_only'] = kwargs.get(
                '_return_http_data_only', True
            )
            kwargs['_preload_content'] = kwargs.get(
                '_preload_content', True
            )
            kwargs['_request_timeout'] = kwargs.get(
                '_request_timeout', None
            )
            kwargs['_check_input_type'] = kwargs.get(
                '_check_input_type', True
            )
            kwargs['_check_return_type'] = kwargs.get(
                '_check_return_type', True
            )
            kwargs['_host_index'] = kwargs.get('_host_index')
            kwargs['deployment_template_id'] = \
                deployment_template_id
            kwargs['deployment_template'] = \
                deployment_template
            return self.call_with_http_info(**kwargs)

        if self.update_deployment_template is None:
            self.update_deployment_template = _Endpoint(
                settings={
                    'response_type': (DeploymentTemplate,),
                    'auth': [
                        'token-valid'
                    ],
                    'endpoint_path': '/v1/deployment_templates/{deployment_template_id}',
                    'operation_id': 'update_deployment_template',
                    'http_method': 'PUT',
                    'servers': None,
                },
                params_map={
                    'all': [
                        'deployment_template_id',
                        'deployment_template',
                    ],
                    'required': [
                        'deployment_template_id',
                        'deployment_template',
                    ],
                    'nullable': [
                    ],
                    'enum': [
                    ],
                    'validation': [
                    ]
                },
                root_map={
                    'validations': {
                    },
                    'allowed_values': {
                    },
                    'openapi_types': {
                        'deployment_template_id':
                            (str,),
                        'deployment_template':
                            (DeploymentTemplate,),
                    },
                    'attribute_map': {
                        'deployment_template_id': 'deployment_template_id',
                    },
                    'location_map': {
                        'deployment_template_id': 'path',
                        'deployment_template': 'body',
                    },
                    'collection_format_map': {
                    }
                },
                headers_map={
                    'accept': [
                        'application/json'
                    ],
                    'content_type': [
                        'application/json'
                    ]
                },
                api_client=api_client,
                callable=__update_deployment_template
            )

    create_deployment = None 
    create_deployment_instance = None 
    create_deployment_template = None 
    delete_deployment = None 
    delete_deployment_instance = None 
    delete_deployment_template = None 
    get_deployment = None 
    get_deployment_instance = None 
    get_deployment_template = None 
    list_deployment_instances = None 
    list_deployment_templates = None 
    list_deployments = None 
    update_deployment = None 
    update_deployment_instance = None 
    update_deployment_template = None 
